@@ -9,17 +9,133 @@
 #include "mediaplaylistproxymodel.h"
 #include "mediaplaylist.h"
 #include "playListLogging.h"
+#include "elisa_settings.h"
+#include "config-upnp-qt.h"
 #include <QItemSelection>
 #include <QList>
-#include <QMediaPlaylist>
 #include <QRandomGenerator>
 #include <QFile>
 #include <QFileInfo>
 #include <QDir>
 #include <QMimeDatabase>
 
+#if KF5KIO_FOUND
+#include <KIO/OpenUrlJob>
+#endif
+
 #include <algorithm>
 
+QList<QUrl> M3uPlaylistParser::fromPlaylist(const QUrl &fileName, const QByteArray &fileContent) {
+    Q_UNUSED(fileName);
+    QList<QUrl> result;
+
+    for (const QByteArray &l : fileContent.split('\n')) {
+        const QString &line = QString::fromUtf8(l);
+        if (!line.isEmpty() && !line.startsWith(QStringLiteral("#"))) {
+            const QUrl &url = line.contains(QStringLiteral("://")) ? QUrl(line) : QUrl::fromLocalFile(line);
+            result.append(url);
+        }
+    }
+
+    return result;
+}
+
+QString M3uPlaylistParser::toPlaylist(const QUrl &fileName, const QList<QString> &listOfUrls) {
+    Q_UNUSED(fileName);
+    QString result;
+
+    for (const QString &line : listOfUrls) {
+        if (!line.isEmpty()) {
+            result += line + QStringLiteral("\n");
+        }
+    }
+
+    return result;
+}
+
+QList<QUrl> PlsPlaylistParser::fromPlaylist(const QUrl &fileName, const QByteArray &fileContent) {
+    Q_UNUSED(fileName);
+    QList<QUrl> result;
+    for (const QByteArray &l : fileContent.split('\n')) {
+        const QString &line = QString::fromUtf8(l);
+        if (!line.isEmpty() && line.startsWith(QStringLiteral("File")) && line.contains(QStringLiteral("="))) {
+            int indexOfEquals = line.indexOf(QStringLiteral("="));
+            QString urlAsString = line.mid(indexOfEquals + QStringLiteral("=").length());
+            const QUrl &url = urlAsString.contains(QStringLiteral("://")) ? QUrl(urlAsString) : QUrl::fromLocalFile(urlAsString);
+            result.append(url);
+        }
+    }
+
+    return result;
+}
+
+QString PlsPlaylistParser::toPlaylist(const QUrl &fileName, const QList<QString> &listOfUrls) {
+    Q_UNUSED(fileName);
+    QString result = QStringLiteral(R"--([playlist]
+
+Version=2
+NumberOfEntries=%1
+)--").arg(listOfUrls.count());
+
+    // Sample:
+/*[playlist]
+
+Version=2
+NumberOfEntries=2
+
+File1=http://example.example:8068
+
+File2=example2.mp3
+*/
+
+    int counter=0;
+    for (const QString &line : listOfUrls) {
+        if (!line.isEmpty()) {
+            QString oneLine = QStringLiteral("\nFile%1=%2\n").arg(counter + 1).arg(line);
+            result += oneLine;
+            counter++;
+        }
+    }
+
+    return result;
+}
+
+QList<QUrl> PlaylistParser::fromPlaylist(const QUrl &fileName, const QByteArray &fileContent) {
+    QList<QUrl> result;
+
+    if (fileName.isValid() && !fileName.isEmpty()) {
+        auto mimeType = QMimeDatabase().mimeTypeForUrl(fileName);
+
+        if (mimeType.inherits(QStringLiteral("audio/x-scpls"))) {
+            PlsPlaylistParser plsPlaylistParser;
+            result = plsPlaylistParser.fromPlaylist(fileName, fileContent);
+        } else if (mimeType.name().contains(QStringLiteral("mpegurl"))) {
+            M3uPlaylistParser m3uPlaylistParser;
+            result = m3uPlaylistParser.fromPlaylist(fileName, fileContent);
+        }
+    }
+
+    return result;
+}
+
+QString PlaylistParser::toPlaylist(const QUrl &fileName, const QList<QString> &listOfUrls) {
+    QString result;
+
+    if (fileName.isValid() && !fileName.isEmpty()) {
+        auto mimeType = QMimeDatabase().mimeTypeForUrl(fileName);
+
+        if (mimeType.inherits(QStringLiteral("audio/x-scpls"))) {
+            PlsPlaylistParser plsPlaylistParser;
+            result = plsPlaylistParser.toPlaylist(fileName, listOfUrls);
+        } else if (mimeType.name().contains(QStringLiteral("mpegurl"))) {
+            M3uPlaylistParser m3uPlaylistParser;
+            result = m3uPlaylistParser.toPlaylist(fileName, listOfUrls);
+        }
+    }
+
+    return result;
+}
+
 class MediaPlayListProxyModelPrivate
 {
 public:
@@ -34,8 +150,6 @@ public:
 
     QPersistentModelIndex mNextTrack;
 
-    QMediaPlaylist mLoadPlaylist;
-
     QList<int> mRandomMapping;
 
     QVariantMap mPersistentSettingsForUndo;
@@ -52,13 +166,15 @@ public:
 
     bool mShufflePlayList = false;
 
+    bool mPartiallyLoaded = false;
+
+    QUrl mLoadedPlayListUrl;
+
 };
 
 MediaPlayListProxyModel::MediaPlayListProxyModel(QObject *parent) : QAbstractProxyModel (parent),
     d(std::make_unique<MediaPlayListProxyModelPrivate>())
 {
-    connect(&d->mLoadPlaylist, &QMediaPlaylist::loaded, this, &MediaPlayListProxyModel::loadPlayListLoaded);
-    connect(&d->mLoadPlaylist, &QMediaPlaylist::loadFailed, this, &MediaPlayListProxyModel::loadPlayListLoadFailed);
     d->mRandomGenerator.seed(static_cast<unsigned int>(QTime::currentTime().msec()));
 }
 
@@ -252,8 +368,15 @@ void MediaPlayListProxyModel::setShufflePlayList(const bool value)
 
                 QModelIndexList from;
                 from.reserve(playListSize);
+                // Adding the current track first if it is not the only one
+                if (playListSize > 1) {
+                    if (currentTrackRow() != 0) {
+                        std::swap(d->mRandomMapping[0], d->mRandomMapping[currentTrackRow()]);
+                    }
+                    from.append(index(d->mRandomMapping.at(0), 0));
+                }
                 // Fisher-Yates algorithm
-                for (int i = 0;  i < playListSize - 1; ++i) {
+                for (int i = 1;  i < playListSize - 1; ++i) {
                     const int swapIndex = d->mRandomGenerator.bounded(i, playListSize);
                     std::swap(d->mRandomMapping[i], d->mRandomMapping[swapIndex]);
                     from.append(index(d->mRandomMapping.at(i), 0));
@@ -688,7 +811,7 @@ void MediaPlayListProxyModel::undoClearPlayList()
 
 void MediaPlayListProxyModel::determineTracks()
 {
-    if (!d->mCurrentTrack.isValid()) {
+    if (!d->mCurrentTrack.isValid() || d->mCurrentPlayListPosition != d->mCurrentTrack.row()) {
         for (int row = 0; row < rowCount(); ++row) {
             auto candidateTrack = index(row, 0);
             const auto type = candidateTrack.data(MediaPlayList::ElementTypeRole).value<ElisaUtils::PlayListEntryType>();
@@ -711,46 +834,81 @@ void MediaPlayListProxyModel::determineTracks()
 
 bool MediaPlayListProxyModel::savePlayList(const QUrl &fileName)
 {
-    QMediaPlaylist savePlaylist;
+    QFile outputFile(fileName.toLocalFile());
+    auto open = outputFile.open(QFile::WriteOnly);
+    if (!open) {
+        return false;
+    }
 
-    for (int i = 0; i < rowCount(); ++i) {
-        if (data(index(i,0), MediaPlayList::IsValidRole).toBool()) {
-            data(index(i,0), MediaPlayList::ResourceRole).toUrl();
-            savePlaylist.addMedia(data(index(i,0), MediaPlayList::ResourceRole).toUrl());
+    QList<QString> listOfFilePaths;
+
+    if (Elisa::ElisaConfiguration::self()->alwaysUseAbsolutePlaylistPaths()) {
+        for (int i = 0; i < rowCount(); ++i) {
+            if (data(index(i,0), MediaPlayList::IsValidRole).toBool()) {
+                data(index(i,0), MediaPlayList::ResourceRole).toUrl();
+                listOfFilePaths.append(data(index(i,0), MediaPlayList::ResourceRole).toUrl().toLocalFile());
+            }
+        }
+    } else {
+        QFileInfo playlistFile(fileName.toLocalFile());
+        QDir dir(playlistFile.absolutePath());
+
+        for (int i = 0; i < rowCount(); ++i) {
+            if (data(index(i,0), MediaPlayList::IsValidRole).toBool()) {
+                QFileInfo musicFile(data(index(i,0), MediaPlayList::ResourceRole).toUrl().toLocalFile());
+
+                if (musicFile.dir().path() == dir.absolutePath()) {
+                    listOfFilePaths.append(data(index(i,0), MediaPlayList::ResourceRole).toUrl().toLocalFile());
+                } else {
+                    listOfFilePaths.append(data(index(i,0), MediaPlayList::ResourceRole).toUrl().toLocalFile());
+                }
+            }
         }
     }
 
-    return savePlaylist.save(fileName, "m3u8");
+    PlaylistParser playlistParser;
+    QString fileContent = playlistParser.toPlaylist(fileName, listOfFilePaths);
+
+    outputFile.write(fileContent.toUtf8());
+
+    return true;
 }
 
 void MediaPlayListProxyModel::loadPlayList(const QUrl &fileName)
 {
-    d->mLoadPlaylist.clear();
-    d->mLoadPlaylist.load(fileName, "m3u8");
+    loadPlayList(fileName, ElisaUtils::ReplacePlayList, ElisaUtils::DoNotTriggerPlay);
 }
 
-void MediaPlayListProxyModel::loadPlayListLoaded()
+void MediaPlayListProxyModel::loadPlayList(const QUrl &fileName,
+                                           ElisaUtils::PlayListEnqueueMode enqueueMode,
+                                           ElisaUtils::PlayListEnqueueTriggerPlay triggerPlay)
 {
-    clearPlayList();
+    resetPartiallyLoaded();
 
-    auto newTracks = DataTypes::EntryDataList{};
-    for (int i = 0; i < d->mLoadPlaylist.mediaCount(); ++i) {
-        newTracks.push_back({{{{DataTypes::ElementTypeRole, ElisaUtils::FileName},
-                               {DataTypes::ResourceRole, d->mLoadPlaylist.media(i).request().url()}}}, {}, {}});
+    QFile inputFile(fileName.toLocalFile());
+    bool open = inputFile.open(QFile::ReadOnly);
+    if (!open) {
+        Q_EMIT playListLoadFailed();
     }
 
-    enqueue(newTracks, ElisaUtils::ReplacePlayList, ElisaUtils::DoNotTriggerPlay);
+    const QByteArray fileContent = inputFile.readAll();
+
+    if (enqueueMode == ElisaUtils::ReplacePlayList) {
+        clearPlayList();
+    }
+
+    d->mLoadedPlayListUrl = fileName;
+
+    DataTypes::EntryDataList newTracks;
+    QSet<QString> processedFiles{QFileInfo(inputFile).canonicalFilePath()};
+    loadLocalPlayList(newTracks, processedFiles, fileName, fileContent);
+
+    enqueue(newTracks, enqueueMode, triggerPlay);
 
     Q_EMIT persistentStateChanged();
-
-    d->mLoadPlaylist.clear();
     Q_EMIT playListLoaded();
-}
-
-void MediaPlayListProxyModel::loadPlayListLoadFailed()
-{
-    d->mLoadPlaylist.clear();
-    Q_EMIT playListLoadFailed();
+    Q_EMIT partiallyLoadedChanged();
+    Q_EMIT canOpenLoadedPlaylistChanged();
 }
 
 QVariantMap MediaPlayListProxyModel::persistentState() const
@@ -801,36 +959,135 @@ void MediaPlayListProxyModel::setPersistentState(const QVariantMap &persistentSt
     Q_EMIT persistentStateChanged();
 }
 
-void MediaPlayListProxyModel::enqueueDirectory(const QUrl &fileName, ElisaUtils::PlayListEntryType databaseIdType,
-                                            ElisaUtils::PlayListEnqueueMode enqueueMode,
-                                            ElisaUtils::PlayListEnqueueTriggerPlay triggerPlay, int depth)
+bool MediaPlayListProxyModel::partiallyLoaded() const
 {
-    if (!fileName.isLocalFile()) return;
-    // clear playlist if required
-    if (enqueueMode == ElisaUtils::ReplacePlayList) {
-        if (rowCount() != 0) {
-            clearPlayList();
+    return d->mPartiallyLoaded;
+}
+
+bool MediaPlayListProxyModel::canOpenLoadedPlaylist() const
+{
+#if !KF5KIO_FOUND
+    return false;
+#endif
+    return d->mLoadedPlayListUrl.isValid() && d->mLoadedPlayListUrl.isLocalFile();
+}
+
+void MediaPlayListProxyModel::openLoadedPlayList()
+{
+#if KF5KIO_FOUND
+    auto job = new KIO::OpenUrlJob(d->mLoadedPlayListUrl, QLatin1String("text/plain"));
+    job->start();
+#endif
+}
+
+void MediaPlayListProxyModel::resetPartiallyLoaded()
+{
+    d->mPartiallyLoaded = false;
+
+    Q_EMIT partiallyLoadedChanged();
+}
+
+void MediaPlayListProxyModel::loadLocalFile(DataTypes::EntryDataList &newTracks, QSet<QString> &processedFiles, const QFileInfo &fileInfo)
+{
+    // protection against recursion
+    auto canonicalFilePath = fileInfo.canonicalFilePath();
+    if (processedFiles.contains(canonicalFilePath)) {
+        return;
+    }
+    processedFiles.insert(canonicalFilePath);
+
+    auto fileUrl = QUrl::fromLocalFile(fileInfo.filePath());
+    auto mimeType = d->mMimeDb.mimeTypeForUrl(fileUrl);
+    if (fileInfo.isDir()) {
+        if (fileInfo.isSymLink()) {
+            return;
+        }
+        loadLocalDirectory(newTracks, processedFiles, fileUrl);
+    } else {
+        if (!mimeType.name().startsWith(QLatin1String("audio/"))) {
+            return;
+        }
+        if (ElisaUtils::isPlayList(mimeType)) {
+            QFile file(fileInfo.filePath());
+            if (!file.open(QIODevice::ReadOnly)) {
+                d->mPartiallyLoaded = true;
+                return;
+            }
+            loadLocalPlayList(newTracks, processedFiles, fileUrl, file.readAll());
+            return;
+        }
+        newTracks.push_back({{{DataTypes::ElementTypeRole, ElisaUtils::FileName}, {DataTypes::ResourceRole, fileUrl}}, {}, {}});
+    }
+}
+
+void MediaPlayListProxyModel::loadLocalPlayList(DataTypes::EntryDataList &newTracks,
+                                                QSet<QString> &processedFiles,
+                                                const QUrl &fileName,
+                                                const QByteArray &fileContent)
+{
+    PlaylistParser playlistParser;
+    QList<QUrl> listOfUrls = playlistParser.fromPlaylist(fileName, fileContent);
+
+    int filtered = filterLocalPlayList(listOfUrls, fileName);
+    if (filtered != 0) {
+        d->mPartiallyLoaded = true;
+    }
+
+    for (const QUrl &oneUrl : qAsConst(listOfUrls)) {
+        if (oneUrl.isLocalFile()) {
+            QFileInfo fileInfo(oneUrl.toLocalFile());
+            loadLocalFile(newTracks, processedFiles, fileInfo);
+        } else {
+            newTracks.push_back({{{{DataTypes::ElementTypeRole, ElisaUtils::FileName}, {DataTypes::ResourceRole, oneUrl}}}, {}, {}});
         }
     }
-    // get contents of directory
-    QDir dirInfo = QDir(fileName.toLocalFile());
-    auto files = dirInfo.entryInfoList(QDir::NoDotAndDotDot | QDir::Readable | QDir::Files | QDir::Dirs, QDir::Name);
-    auto newFiles = DataTypes::EntryDataList();
-    for (const auto &file : files)
-    {
-        auto fileUrl = QUrl::fromLocalFile(file.filePath());
-        if (file.isFile() && d->mMimeDb.mimeTypeForUrl(fileUrl).name().startsWith(QLatin1String("audio/")))
-        {
-            newFiles.push_back({{{DataTypes::ElementTypeRole, ElisaUtils::Track},
-                                 {DataTypes::ResourceRole, fileUrl}},{}, {}});
+}
+
+void MediaPlayListProxyModel::loadLocalDirectory(DataTypes::EntryDataList &newTracks, QSet<QString> &processedFiles, const QUrl &dirName)
+{
+    QDir dirInfo(dirName.toLocalFile());
+    auto fileInfoList = dirInfo.entryInfoList(QDir::NoDotAndDotDot | QDir::Readable | QDir::Files | QDir::Dirs, QDir::Name);
+
+    for (const auto &fileInfo : fileInfoList) {
+        loadLocalFile(newTracks, processedFiles, fileInfo);
+    }
+}
+
+int MediaPlayListProxyModel::filterLocalPlayList(QList<QUrl>& result, const QUrl &playlistUrl)
+{
+    int filtered = 0;
+
+    for (auto iterator = result.begin(); iterator != result.end();) {
+        bool exists = true;
+
+        auto &url = *iterator;
+        if (url.isLocalFile()) {
+            QString file = url.toLocalFile();
+
+            QFileInfo fileInfo(file);
+            if (playlistUrl.isLocalFile() && fileInfo.isRelative()) {
+                auto absoluteDir = QFileInfo(playlistUrl.toLocalFile()).absoluteDir();
+                if (fileInfo.isDir()) {
+                    file = absoluteDir.absolutePath() + QDir::separator() + fileInfo.path();
+                } else {
+                    file = absoluteDir.absoluteFilePath(file);
+                }
+                fileInfo.setFile(file);
+                url = QUrl::fromLocalFile(file);
+            }
+
+            exists = fileInfo.exists();
         }
-        else if (file.isDir() && depth > 1)
-        {
-            // recurse through directory
-            enqueueDirectory(fileUrl, databaseIdType, ElisaUtils::AppendPlayList, triggerPlay, depth-1);
+
+        if (exists) {
+            ++iterator;
+        } else {
+            ++filtered;
+            iterator = result.erase(iterator);
         }
     }
-    if (newFiles.size() != 0) enqueue(newFiles, ElisaUtils::AppendPlayList, triggerPlay);
+
+    return filtered;
 }
 
 #include "moc_mediaplaylistproxymodel.cpp"
